/*
 * generate.c
 *
 * Interface between input, and output routines.  Called by main loop.
 *
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97 Akim Demaille, Miguel Santana
 * $Id: generate.c,v 1.14 1997/11/07 13:28:05 demaille Exp $
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * $Id: generate.c,v 1.14 1997/11/07 13:28:05 demaille Exp $
 */

#include "a2ps.h"
#include "generate.h"
#include "psgen.h"
#include "sshread.h"
#include "read.h"
#include "routines.h"
#include "select.h"
#include "darray.h"
#include "delegate.h"
#include "gen.h"
#include "msg.h"
#include "printers.h"
#include "metaseq.h"

extern struct a2ps_job * job;
extern int delegate;
extern int highlight_level;
extern char * style_request;
extern int isdir PARAMS ((const char *path));
extern enum eol end_of_line;

/************************************************************************/
/*			The inputs					*/
/************************************************************************/
static struct buffer *
new_input (uchar * name)
{
  struct buffer * res;
  FILE * input_stream;
  struct file_job * file_job;
  struct stat statbuf;		/* to get file modification time */
  struct tm *tm;

  a2ps_open_input_session (job, name);
  file_job = CURRENT_FILE (job);

  /* Retrieve file modification date and hour */
  if (!IS_EMPTY(name)) {
    /* Printing a file given by its path */
    if (isdir ((char *) name)) {
	error (0, 0, _("`%s' is a directory"), name);
	file_job->printable = FALSE;
    }
    
    file_job->name = (uchar *) name;
    if ((input_stream = fopen ((char *) name, "r")) == NULL) {
      error (0, errno, _("cannot open file `%s'"), name);
      file_job->printable = FALSE;
    } else if (stat ((char *) name, &statbuf) == -1) {
      error (0, errno, _("cannot get informations on file `%s'"), 
	     name);
      file_job->printable = FALSE;
    } else { 
      time_t tim = statbuf.st_mtime;
      tm = localtime (&tim);
      memcpy ((POINTER) &(file_job->mod_tm),
	      (POINTER) tm, sizeof (*tm));
    }
  }
  else
    input_stream = stdin;

  /*
   * What should be done out of this file?
   * Find the command associated to that file
   * - UNPRINTABLE for unprintable
   * - requested style sheet key
   * - style sheet key
   */
  if (!file_job->printable)
    file_job->type = "UNPRINTABLE";
  else if (IS_EMPTY (style_request))
    file_job->type = get_command ((char *) file_job->name);
  else
    file_job->type = style_request;
  
  res = buffer_stream_new (input_stream, end_of_line);
  return res;
}

static void
end_input (struct buffer * buffer)
{
  if (buffer->stream && buffer->stream!= stdin)
    fclose (buffer->stream);
  buffer_free (buffer);

  a2ps_close_input_session (job);
}

/************************************************************************/
/*			The producers					*/
/************************************************************************/
/*
 * Make on message on what we did
 */
static void
message_file_pages_printed (a2ps_job * Job, const char * stylename)
{
  int sheets;
  
  sheets = CURRENT_FILE (Job)->sheets;
  if (Job->rectoverso)
    sheets = (sheets + 1) / 2;
  
  if (CURRENT_FILE (Job)->pages == 1) 
    /* 1 page on 1 sheet */
    message (msg_report2, _("[%s (%s): 1 page on 1 sheet]\n"), 
	     CURRENT_FILE (Job)->name, 
	     stylename);
  else if (sheets == 1)
    /* several pages on 1 sheet */
    message (msg_report2, _("[%s (%s): %d pages on 1 sheet]\n"), 
	     CURRENT_FILE (Job)->name,
	     stylename,
	     CURRENT_FILE (Job)->pages);
  else
    /* several sheets */
    message (msg_report2, _("[%s (%s): %d pages on %d sheets]\n"), 
	     CURRENT_FILE (Job)->name,
	     stylename,
	     CURRENT_FILE (Job)-> pages,
	     sheets);
}

/*
 * Total printed
 */
void
message_job_pages_printed (a2ps_job * Job)
{
  int sheets;
  uchar buf [256];
  
  sheets = Job->sheets;
  if (Job->rectoverso)
    sheets = (sheets + 1) / 2;
  
  /* Make a nice message to tell where the output is sent */
  final_destination_to_string (Job, buf);
  
  /* Report the pages */
  if (Job->pages == 1) 
    /* 1 page on 1 sheet "sent to the default printer" etc. */
    message (msg_report1, _("[Total: 1 page on 1 sheet] %s\n"), buf);
  else if (sheets == 1)
    /* several pages on 1 sheet */
    message (msg_report1, _("[Total: %d pages on 1 sheet] %s\n"), 
	     Job->pages, buf);
  else
    /* several sheets */
    message (msg_report1, _("[Total: %d pages on %d sheets] %s\n"), 
	     Job-> pages, sheets, buf);
}
/*
 * Total printed
 */
void
message_nothing_printed (void)
{
  message (msg_report1, _("[No output produced]\n"));
}

void
print_toc (const uchar * name, const uchar * value, int * native_jobs)
{
  struct buffer * toc_buffer;
  uchar * toc_content;

  /* Create a entry for the toc, as if it were a regular file */
  a2ps_open_input_session (job, xustrdup (name));
  /* But it is not a regular file: we need to be able to know
   * that it is indeed a toc, so that --pages=toc can be honored */
  CURRENT_FILE (job)->is_toc = TRUE;

  austrcpy (toc_content,
	    expand_user_string (job, CURRENT_FILE (job),
				name, value));
  toc_buffer = buffer_string_new (toc_content, end_of_line);
  
  /* We typeset it with PreScript */
  ssh_print_postscript (job, toc_buffer, get_style_sheet ("pre"));
  (*native_jobs)++;
  
  a2ps_close_input_session (job);
  buffer_free (toc_buffer);
}

/*
 * Called by the main loop. 
 * The file to print is the last of the darray job->jobs.
 * Return TRUE if was a success, FALSE otherwise
 */
void
print (uchar * filename, int * native_jobs, int * delegated_jobs)
{
  char buf[512];
  struct delegation * contract = NULL;
  struct style_sheet * sheet;
  struct buffer * input_buffer;
  style_kind_t style_kind;
  
  /*
   * First, open that file and get info about it
   * It may seem useless in some cases (e.g. the file will be subcontracted)
   * but it ensures that readbility, and stat can be correctly done.
   */
  input_buffer = new_input (filename);

  if (delegate
      && (contract = 
	  get_subcontract (CURRENT_FILE (job)->type,
			   output_format_to_key (job->output_format))))
    style_kind = subcontracted;
  else
    style_kind = string_to_style_kind (CURRENT_FILE (job)->type);
  
  message (msg_file1,
	   "Getting ready to print file `%s', with command `%s'\n",
	   CURRENT_FILE (job)->name, CURRENT_FILE (job)->type);

  /*
   * Then do it
   */
  switch (style_kind) {
  case subcontracted:
    /* In ps generation, we must begin a new page */
    require_fresh_page (job);
    sprintf (buf, _("%s, delegated to %s"), 
	     CURRENT_FILE (job)->type, contract->name);
    if (subcontract (CURRENT_FILE (job), contract)) {
      (*delegated_jobs)++;
      message_file_pages_printed (job, buf);
    } else
      message (msg_report2, _("[%s (%s): failed.  Ignored]\n"), 
	       CURRENT_FILE (job)->name, buf);
    break;

  case unprintable:
    /* The job will not be processed correctly */
    message (msg_report2,
	     _("[%s (unprintable): ignored]\n"), CURRENT_FILE (job)->name);
    break;

  case binary:
    if (job->print_binaries)
      goto plain_print;

    message (msg_report2,
	     _("[%s (binary): ignored]\n"), CURRENT_FILE (job)->name);
    break;

  case sshparser:
    /* If highlight_level == none, don't */
    if (highlight_level == 0)
      goto plain_print;
    sheet = get_style_sheet (CURRENT_FILE (job)->type);
    if (!sheet)
      goto plain_print;

    buffer_set_lower_case (input_buffer,
			   sheet->sensitiveness == CASE_INSENSITIVE);
    ssh_print_postscript (job, input_buffer, sheet);
    message_file_pages_printed (job, (char *) sheet->name);
    (*native_jobs)++;
    break;
    
  plain_print:
  case no_style:
    plain_print_postscript (job, input_buffer);
    message_file_pages_printed (job, _("plain"));
    (*native_jobs)++;
    break;
  }

  end_input (input_buffer);
}
