%{ /* -*- c -*- */
/*
 * Grammar for parsing the style sheets
 *
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97 Akim Demaille, Miguel Santana
 *
 */

/*
 * This file is part of a2ps
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * $Id: parsessh.y,v 1.53 1997/11/07 13:28:12 demaille Exp $
 */

#include "a2ps.h"
#include "jobs.h"
#include "faces.h"
#include "ssheet.h"
#include "msg.h"
#include "routines.h"
#include "yy2ssh.h"
#include "regex.h"

#define YYDEBUG 1
#define YYERROR_VERBOSE 1
#define YYPRINT(file, type, value)   yyprint (file, type, value)

/* We need to use the same `const' as bison, to avoid the following
 * prototypes to diverge from the function declarartions */
#undef const
#ifndef __cplusplus
#ifndef __STDC__
#define const
#endif
#endif

/* Comes from the parser */
extern int sshlineno;

/* Comes from the caller */
extern FILE * sshin;
extern struct a2ps_job * job;
const char * yyfilename;

/* Local prototypes */
void yyerror PARAMS ((const char *msg));
static void yyprint ();

/* Initilizes the obstacks */
void sshlex_initialize PARAMS ((void));

/* Comes from main.c */
extern int highlight_level;

int yylex PARAMS ((void));
int parse_style_sheet PARAMS ((const char * filename));

%}
%union
{
  int integer;
  uchar * string;
  struct pattern * pattern;
  struct style_sheet * sheet;
  struct symbol * symbol;
  struct sequence * sequence;
  struct darray * array;
  struct faced_string * faced_string;
  face_t face;
  CASE_SENSITIVENESS sensitiveness;
}

%token <string> tLATEXSYMBOL
%token <face> tFACE
%token <string> tSYMBOL tSTRING
%token <pattern> tREGEX
%token tSTYLE tIS tEND tKeywordS tARE tIN tOPERATORS tSEQUENCES
%token tFIRST tSECOND tALPHABET tALPHABETS tDOCUMENTATION tEXCEPTIONS
%token tCASE tCSTRING tCCHAR tOPTIONAL tCLOSERS
%token tWRITTEN tBY tVERSION tREQUIRES tA2PS tANCESTORS
%token <sensitiveness> tSENSITIVENESS
%token <integer> tBACK_REF

%type <face> face face_exp
%type <string> word comment
%type <string> authors version long_string requirement
%type <faced_string> a_destination
%type <symbol> symbol 
%type <symbol> keywords_symbol keyword_regex
%type <symbol> operators_symbol operator_regex
%type <sheet> style_sheet definition_list
%type <array> keywords_def keywords_symbol_list
%type <array> operators_def operators_symbol_list
%type <array> sequence_list sequence_def
%type <array> exception_def_opt
%type <array> ancestors_def ancestors_list
%type <array> closers_opt 
%type <sensitiveness> case_def
%type <array> destination destination_list
%type <sequence> sequence
%%

/************************************************************************/
/*		Top most						*/
/************************************************************************/
file	:
  style_sheet
    { 
      char * cp, *cp2;
      
      /* Look for the key of the style, 
       * which is the name of the file */
      cp = strrchr (yyfilename, DIR_SEP);
      if (cp)
	cp2 = xstrdup (cp + 1);
      else
	cp2 = xstrdup (yyfilename);
      /* No need of the suffix */
      *strrchr (cp2, '.') = '\0';
      $1->key = xstrdup (cp2);
      FREE (cp2);
      
      /* Do not finalize the style sheet here, because
       * finalizing may call the parser a second time, and
       * it is *not* reentrant */
      add_style_sheet ($1);
      
      if (message_test (msg_sheet3))
	style_sheet_self_print (stderr, $1);
    }
  ;

style_sheet :
   tSTYLE word tIS definition_list tEND tSTYLE
      { 
	$4->name = $2;
	$$ = $4;
      }
   ;

definition_list : 
	/* empty */ { 
	  $$ = new_style_sheet ((const uchar *) "<no name>"); 
	}
	| definition_list tOPTIONAL tKeywordS keywords_def tKeywordS
	{
	  if (highlight_level == 2) {
	    symbols_set_no_faces ($4, Plain);
	    da_qsort ($4);
	    words_merge_symbols_unique ($1->keywords, $4);
	    da_erase ($4);
	  } else
	    da_free ($4, (da_map_func_t) free_symbol);
	  $$ = $1;
	}
	| definition_list tKeywordS keywords_def tKeywordS
	{
	  symbols_set_no_faces ($3, Plain);
	  da_qsort ($3);
	  words_merge_symbols_unique ($1->keywords, $3);
	  da_erase ($3);
	  $$ = $1;
	}
	| definition_list tOPTIONAL tOPERATORS operators_def tOPERATORS
	{
	  if (highlight_level == 2) {
	    symbols_set_no_faces ($4, Plain);
	    da_qsort ($4);
	    words_merge_symbols_unique ($1->operators, $4);
	    da_erase ($4);
	  } else
	    da_free ($4, (da_map_func_t) free_symbol);
	  $$ = $1;
	}
	| definition_list tOPERATORS operators_def tOPERATORS
	{
	  symbols_set_no_faces ($3, Plain);
	  da_qsort ($3);
	  words_merge_symbols_unique ($1->operators, $3);
	  da_erase ($3);
	  $$ = $1;
	}
	| definition_list tOPTIONAL sequence_def {
	  if (highlight_level == 2) {
	    da_concat ($1->sequences, $3);
	    da_erase ($3);
	  } else
	    da_free ($3, (da_map_func_t) free_sequence);
	  $$ = $1;
	}
	| definition_list sequence_def {
	  da_concat ($1->sequences, $2);
	  da_erase ($2);
	  $$ = $1;
	}
	| definition_list ancestors_def {
	  da_concat ($1->ancestors, $2);
	  da_erase ($2);
	  $$ = $1;
	}
	| definition_list tALPHABETS tARE word {
	  string_to_array ($1->alpha1, $4);
	  string_to_array ($1->alpha2, $4);
	  /* This is the syntax table used by regex */
	  FREE ($4);
	  $4 = NULL;
	  $$ = $1;
	}
	| definition_list tFIRST tALPHABET tIS word {
	  string_to_array ($1->alpha1, $5);
	  /* This is the syntax table used by regex */
	  FREE ($5);
	  $5 = NULL;
	  $$ = $1;
	}
	| definition_list tSECOND tALPHABET tIS word {
	  string_to_array ($1->alpha2, $5);
	  /* This is the syntax table used by regex */
	  FREE ($5);
	  $5 = NULL;
	  $$ = $1;
	}
	| definition_list case_def {
	  $1->sensitiveness = $2;
	}
	| definition_list comment {
	  $1->comment = $2;
	}
	| definition_list authors {
	  $1->author = $2;
	}
	| definition_list version {
	  style_sheet_set_version ($1, (const char *) $2);
	}
	| definition_list requirement { 
	  /* Make sure now that we won't encounter new tokens.
	   * This avoids nasty error messages, or worse:
	   * unexpected behavior at run time */
	  if (!style_sheet_set_requirement ($1, (const char *) $2))
	    error (1, 0, 
		   _("cannot process `%s' which requires a2ps version %s"),
		   yyfilename, $2);
	}
	;

/************************************************************************/
/*		Dealing with the inessential informations		*/
/************************************************************************/
requirement : 
	  tREQUIRES tA2PS tSYMBOL { $$ = $3 ; } ;
	| tREQUIRES tA2PS tVERSION tSYMBOL { $$ = $4 ; } ;

comment : tDOCUMENTATION tIS long_string tEND tDOCUMENTATION { $$ = $3 ; };

long_string: word { $$ = $1; }
	| long_string word
	{
	  size_t len1;
	  size_t len2;
	  
	  len1 = ustrlen ($1);
	  $1[len1] = '\n';
	  len2 = ustrlen ($2);
	  $$ = ALLOC (uchar, len1 + len2 + 2);
	  ustpcpy (ustpncpy ($$, $1, len1 + 1), $2);
	  FREE ($1);
	  FREE ($2);
	}
	;

authors : tWRITTEN tBY word	{ $$ = $3 ; };

version : 
	  tVERSION tIS word 	{ $$ = $3 ; };
	| tVERSION word 	{ $$ = $2 ; };

/************************************************************************/
/*		Dealing with the ancestors of a style sheet		*/
/************************************************************************/
ancestors_def : tANCESTORS tARE ancestors_list tEND tANCESTORS 
	{ 
	  /* The list of keys of style sheets from which it
	   * inherits */
  	  $$ = $3; 
	}
	;

ancestors_list: word 
	{ 
	  /* Create a list of ancestors, and drop the new
	   * one in */
	  $$ = da_new ("Ancestors tmp", 2, 
		       da_linear, 2,
		       (da_print_func_t) da_str_print, NULL);
	  da_append ($$, $1);
	} 
	| ancestors_list ',' word 
	{
	  da_append ($1, $3); 
	  $$ = $1;
	}
	;

case_def : tCASE tSENSITIVENESS { $$ = $2 ; } ;

/************************************************************************/
/*	Destinations							*/
/* (Lists of (strings/regexp back references, face)			*/
/************************************************************************/
destination:
	  a_destination
	{ 
	  $$ = destinations_new ();
	  destinations_add ($$, $1);
	}
	;

a_destination:
	  word face	
	{ 
	  $$ = faced_string_new  ($1, 0, $2);
	}
	| word
	{ 
	  $$ = faced_string_new  ($1, 0, No_face);
	}
	| face
	{ 
	  $$ = faced_string_new  (NULL, 0, $1);
	}
	| tBACK_REF face
	{ 
	  $$ = faced_string_new  (UNULL, $1, $2);
	}
	| tBACK_REF
	{ 
	  $$ = faced_string_new  (UNULL, $1, No_face);
	}
	| tLATEXSYMBOL
	{
	  $$ = faced_string_new ($1, 0, Symbol);
	}
	;

destination_list:
	a_destination
	{
	  $$ = destinations_new ();
	  destinations_add ($$, $1);
	}
	|
	destination_list ',' a_destination
	{
	  destinations_add ($1, $3);
	  $$ = $1;
	}
	;

face:	tFACE 			{ $$ = $1; }
	| '(' face_exp ')'	{ $$ = $2; }
	;

face_exp: tFACE 		{ $$ = $1; }
	| face_exp '+' tFACE 	{ $$ = $1 | $3;}
	;

/************************************************************************/
/*		Symbol atoms						*/
/************************************************************************/
/* A symbol is the group of 1. string to match, 2. the destination
 * 
 * In the special shortcut where no face is given, see where
 * symbol_list appears for resolving */
symbol:   word destination
	{
	  $$ = new_symbol ($1, NULL, $2);	
	}
	| word
	{
	  $$ = new_symbol ($1, NULL,
			   destinations_new_single (UNULL, 0, No_face));
	}
	| '(' word destination_list ')'
	{
	  $$ = new_symbol ($2, NULL, $3);
	}
	;

word:	tSYMBOL { $$ = $1;}
	| tSTRING { $$ = $1;}
	;

/************************************************************************/
/*		Keywords lists						*/
/* We make the difference because the regex must be compiled with a	*/
/* difference (which is \\b\\(%s\\)\\b) between keywords and operators	*/
/************************************************************************/
keywords_def : tIN face tARE keywords_symbol_list tEND { 
	  symbols_set_no_faces ($4, $2);
	  $$ = $4;
	}
	| tARE keywords_symbol_list tEND { 
  	  /* First of all, the No_face must be turned into Plain */
  	  $$ = $2;
	}
	;

keywords_symbol_list: keywords_symbol
	{
	  $$ = da_new ("A list of symbols for keywords", 100,
		       da_linear, 100,
		       (da_print_func_t) symbol_self_print, 
		       (da_cmp_func_t) symbol_cmp);
	  da_append ($$, $1);
	} 
	| keywords_symbol_list ',' keywords_symbol
	{
	  da_append ($1, $3); 
	  $$ = $1;
	}
	;

/* A symbol is the group of 1. string to match, 2. string to
 * print, 3. face of the string to print.  In the special
 * shortcut where no face is given, see where symbol_list appears
 * for resolving */
keywords_symbol:   
	  symbol
	{ 
	  $$ = $1 ;
	}
	| keyword_regex
	{
	  $$ = $1;
	}
	;

keyword_regex:
	  tREGEX destination
	{ 
	  $$ = new_keyword_symbol (UNULL, $1, $2);
	}
	| tREGEX
	{
	  $$ = new_keyword_symbol (UNULL, $1,
				   destinations_new_single (UNULL, 0, No_face));
	}
	| '(' tREGEX destination_list ')'
	{
	  $$ = new_keyword_symbol (UNULL, $2, $3);
	}
	;

/************************************************************************/
/*		Operators lists						*/
/************************************************************************/
operators_def: tIN face tARE operators_symbol_list tEND { 
	  symbols_set_no_faces ($4, $2);
	  $$ = $4;
	}
	| tARE operators_symbol_list tEND { 
  	  /* First of all, the No_face must be turned into Plain */
  	  $$ = $2;
	}
	;

operators_symbol_list: operators_symbol
	{ 
	  $$ = da_new ("A list of symbols", 100,
		       da_linear, 100,
		       (da_print_func_t) symbol_self_print, 
		       (da_cmp_func_t) symbol_cmp);
	  da_append ($$, $1);
	} 
	| operators_symbol_list ',' operators_symbol
	{
	  da_append ($1, $3); 
	  $$ = $1;
	}
	;

/* A symbol is the group of 1. string to match, 2. string to
 * print, 3. face of the string to print.  In the special
 * shortcut where no face is given, see where symbol_list appears
 * for resolving */
operators_symbol:  
	symbol
	{
	  $$ = $1 ;
	}
	|
	operator_regex
	{
	  $$ = $1;
	}

operator_regex:
	  tREGEX destination
	{
	  $$ = new_symbol (UNULL, $1, $2);
	}
	| tREGEX
	{
	  $$ = new_symbol (UNULL, $1,
			   destinations_new_single (UNULL, 0, No_face));
	}
	| '(' tREGEX destination_list ')'
	{
	  $$ = new_symbol (UNULL, $2, $3);
	}
	;

/************************************************************************/
/*		Dealing with the sequences				*/
/************************************************************************/
sequence_def : tSEQUENCES tARE sequence_list tEND tSEQUENCES { $$ = $3; }
	;

sequence_list: sequence { 
	  $$ = da_new ("Sequence tmp", 100, 
		       da_linear, 100,
		       (da_print_func_t) sequence_self_print, NULL);
	  da_append ($$, $1);
	} 
	| sequence_list ',' sequence {
	  da_append ($1, $3); 
	  $$ = $1;
	}
	;

/*
 * I can understand one will get sick reading this.  It hurts, indeed,
 * but is necessary because of shift/reduce conflicts if one uses
 * 
 *    sequence: operators_symbol face closers_opt exception_def_opt
 * 
 * The problem is that when bison/yacc reads this
 *
 *    <word> <word> . <face1> <face2>
 *
 * it doesn't know whether <face1> should be in $1, or in $2.
 * So we have to inline the rule at hand...
 */
sequence: 
	/* Expansion of each possibility for
	 * <operators_symbol> <face> <closers_opt> <exception_def_opt>
	 */
	word tLATEXSYMBOL face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol ($1, NULL,
				    destinations_new_single ($2, 0, Symbol));
	  $$ = new_sequence (open_symbol, $3, $4, $5);
	}
	| word word face face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol ($1, NULL,
				    destinations_new_single ($2, 0, $3));
	  $$ = new_sequence (open_symbol, $4, $5, $6);
	}
	| word face face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol ($1, NULL,
				    destinations_new_single (UNULL, 0, $2)); 
	  $$ = new_sequence (open_symbol, $3, $4, $5);
	}
	| word word face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol ($1, NULL,
				    destinations_new_single ($2, 0, $3)); 
	  $$ = new_sequence (open_symbol, $3, $4, $5);
	}
	| word face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol ($1, NULL,
				    destinations_new_single (UNULL, 0, $2));
	  $$ = new_sequence (open_symbol, $2, $3, $4);
	}
	| tREGEX word face face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol (UNULL, $1,
				    destinations_new_single ($2, 0, $3));
	  $$ = new_sequence (open_symbol, $4, $5, $6);
	}
	| tREGEX face face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol (UNULL, $1,
				    destinations_new_single (UNULL, 0, $2));
	  $$ = new_sequence (open_symbol, $3, $4, $5);
	}
	| tREGEX word face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol (UNULL, $1,
				    destinations_new_single ($2, 0, $3));
	  $$ = new_sequence (open_symbol, $3, $4, $5);
	}
	| tREGEX face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol (UNULL, $1,
				    destinations_new_single (UNULL, 0, No_face));
	  $$ = new_sequence (open_symbol, $2, $3, $4);
	}
	| '(' word destination_list ')' face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol ($2, NULL, $3);
	  $$ = new_sequence (open_symbol, $5, $6, $7);
	}
	| '(' tREGEX destination_list ')' face closers_opt exception_def_opt
	{
	  struct symbol * open_symbol;
	  open_symbol = new_symbol (UNULL, $2, $3);
	  $$ = new_sequence (open_symbol, $5, $6, $7);
	}
/* End of the brute force expansion */
	| tCSTRING
	{ 
	  $$ = new_C_string_sequence ("\"");
	}
	| tCCHAR
	{
	  $$ = new_C_string_sequence ("\'");
	}
	;

closers_opt: 
	/* Nothing */
	{
	  /* This is a shortcut which means "up to the end of the line" */
	  struct pattern * pattern;
	  
	  $$ = da_new ("Closing alternatives", 2, da_linear, 2,
			  (da_print_func_t) symbol_self_print, NULL);
	  pattern = new_pattern (xstrdup ("$"), 1);
	  da_append ($$, new_symbol (NULL, pattern,
				     destinations_new_single (NULL, 0, No_face)));
	}
	| operators_symbol
	{
	  /* Only one */
	  $$ = da_new ("Closing alternatives", 5,
		       da_linear, 5,
		       (da_print_func_t) symbol_self_print, NULL);
	  da_append ($$, $1);
	};
	| tCLOSERS operators_def tCLOSERS
	{ 
	  /* Several, comma separated, between () */
	  $$ = $2; 
	}
	;

exception_def_opt: /* Nothing */
	{ 
	  $$ = da_new ("Exceptions", 1,
		       da_linear, 1,
		       (da_print_func_t) symbol_self_print,
		       (da_cmp_func_t) symbol_cmp);
	}
	| tEXCEPTIONS operators_def tEXCEPTIONS
	{ 
	  $$ = $2;
	};

%%

void
yyerror (const char *msg)
{
  error_at_line (1, 0, yyfilename, sshlineno, msg);
}

/*
 * FIXME: Cover this other types
 */
static void
yyprint (FILE *file, int type, YYSTYPE value)
{
  switch (type) {
  case tFACE:
    fprintf (file, " %s", face_to_string (value.face));
    break;
  case tSYMBOL:
    fprintf (file, " %s", value.string);
    break;
  case tSTRING:
    fprintf (file, " `%s'", value.string);
    break;
  }
}

int
parse_style_sheet (const char * filename)
{
  int res;

  yyfilename = filename;
  sshlineno = 1;
  sshin = xrfopen (yyfilename);

  message (msg_file1 | msg_sheet1, "Parsing file `%s'\n", yyfilename);

  sshlex_initialize ();

  if (message_test (msg_sheet3))
    yydebug = TRUE;
  res = yyparse ();

  fclose (sshin);
  return res;
}
