/*
 * ssheet.h
 *
 * definition of the languages style sheets
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97 Akim Demaille, Miguel Santana
 * $Id: ssheet.h,v 1.44 1997/10/31 18:06:17 demaille Exp $
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _SSHEET_H_
#define _SSHEET_H_

#include "a2ps.h"
#include "darray.h"
#include "faces.h"
#include "hashtab.h"
#include "jobs.h"
#include "regex.h"

typedef enum { 
  CASE_SENSITIVE, CASE_INSENSITIVE
} CASE_SENSITIVENESS; 

/*
 * Used by the scanner and parser to return a pattern,
 * which may enclose NUL
 */
struct pattern {
  char * pattern;
  size_t len;
};

/*
 * Now, keywords, and operators share the same underlying
 * structure, because, though keywords don't need a SYMBOL field,
 * they are so close, that implementation should be the same.
 * This may be a sign that OO would have been appreciated :)
 */
struct faced_string {
  uchar * string;		/* the destination string	*/
  int reg_ref;			/* there is no destination string,
				 * but the model is the REG_REF expression
				 * caught by the regexp (e.g. 1 for \1 */
  face_t face;
};

struct symbol {
  uchar * word;
  struct re_pattern_buffer * regex;
  struct darray * destination;
};

struct words {
  /* darrays of struct symbol * */
  struct darray * strings;	/* those which matchers are strings */
  struct darray * regexps;	/* those which matchers are regexps */
  struct symbol ** min [256];
  struct symbol ** max [256];
};

struct sequence {
  struct symbol * open;
  face_t face;
  struct darray * exceptions;		/* Exception, i.e. \" between ""  */
  struct darray * close;		/* closing alternatives */
} ; 

struct style_sheet {
  /* index of the language (compare with command-line option) */
  const char * key;

  /* Nice looking name of the style */
  const uchar * name;

  /* Who wrote it, and when */
  const uchar * author;
  int version[4];

  /* What version of a2ps is required? */
  int requirement[4];

  /* Note describing the mode or the language */
  const uchar * comment;
  
  /* Does it have ancestors (i.e., this one is an extension of
   * its ancestors).  It is a list of keys, of course */
  struct darray * ancestors;

  /* case sensitiveness for keywords and symbols */
  CASE_SENSITIVENESS sensitiveness;

  /* definition of the "words" (keywords and symbols):
   * a char belonging to a first alphabet (first_characters),
   * and any number of chars belonging to the second (following_characters).
   * The definition of those alphabet is alphanumerical plus 
   * ascii intervals */

  /* First characters alphabet:
     begin_char_1, end_char_1,	ie ascii code between char 1 & char 2
     begin_char_2, end_char_2, ...
     \0 */
  uchar alpha1 [256];

  /* the other than first letter may be: */
  uchar alpha2 [256];

  /* list of keywords for this language
     keyword_1, font_1,
     keyword_2, font_2, ...*/
  struct words * keywords;

  /* same as symbols BUT there is no need to be preceded and followed
     by not in extended alphabet characters. In other words, these are
     not "words" of the alphabet, but any sequence of chars */
  struct words * operators;

  /* type of sequences (strings, comments, etc.) :
     begin_string_1, font_for_begin_string_1, font_for_inside_sequence_1, 
     	end_string_1, font_for_end_string_1, 
     begin_string_2, font_for_begin_string_2, font_for_inside_sequence_2, 
     	end_string_2, font_for_end_string_2, ...
     \0 */
  struct darray * sequences;
};

/*	Notes
 *
 * - global exceptions is made for languages such as tcl where, whereever the
 * symbol appears, it must not be "understood", e.g. \" and \\ are
 * written directly.
 *
 * - local exception is made for ! languages such as ada where "" is the
 * empty string, but appearing in a string, "" denotes ".
 */


/*
 * The faced_string
 */
struct faced_string *
faced_string_new PARAMS ((uchar * string, int reg_ref, face_t face));

/*
 * The version numbers
 */
void style_sheet_set_version PARAMS ((struct style_sheet * sheet,
				      const char * version_string));
int style_sheet_set_requirement PARAMS ((struct style_sheet * sheet,
					  const char * version_string));

/*
 * The destinations (part of symbol)
 */
struct darray * destinations_new PARAMS ((void));
struct darray * 
destinations_new_single PARAMS ((uchar * string, int reg_ref, face_t face));
void
destinations_add PARAMS ((struct darray * dest, struct faced_string * str));
void
destinations_self_print PARAMS ((FILE * stream, struct darray * destinations));

/*
 * Dealing with the keywords, symbols and operators
 */
struct symbol * new_symbol PARAMS ((uchar * word,
				    struct pattern * pattern,
				    struct darray * destination));
struct symbol * new_keyword_symbol PARAMS ((uchar * word,
					    struct pattern * pattern,
					    struct darray * destination));
void free_symbol PARAMS ((struct symbol * symbol));
int symbol_cmp PARAMS ((struct symbol * k1, struct symbol * k2));
void symbol_self_print PARAMS ((FILE * stream, struct symbol * k));
void symbol_set_no_face PARAMS ((struct symbol * symbol, face_t face));
void symbols_set_no_faces PARAMS ((struct darray * arr, face_t face));

/*
 * Dealing with the sequences
 */
struct pattern * 
new_pattern PARAMS ((char * pattern, size_t len));

struct sequence * new_sequence 
   	PARAMS ((struct symbol * Open,
		 face_t in_face,
		 struct darray * Close,
		 struct darray * exceptions));
void free_sequence PARAMS ((struct sequence * sequence));
struct sequence * new_C_string_sequence PARAMS ((const char * delimitor));
void sequence_self_print PARAMS ((FILE * stream, struct sequence * v));


/*
 * Dealing with the struct words
 */
void words_merge_symbols_unique PARAMS ((struct words * words, 
					 struct darray * new));

/*
 * Dealing with the style sheets
 */
struct style_sheet * new_style_sheet PARAMS ((const uchar * name));
void style_sheet_finalize PARAMS ((struct style_sheet * sheet));
void style_sheet_self_print PARAMS ((FILE * stream,
				     struct style_sheet * sheet));

/*
 * Dealing with the hash table for style sheets
 */
struct hash_table * new_style_sheets PARAMS ((void));
void require_style_sheet PARAMS ((const char * key));
struct style_sheet * get_style_sheet PARAMS ((const char * name));
void add_style_sheet PARAMS ((struct style_sheet * sheet));

/************************************************************************/
/*				style selection				*/
/************************************************************************/
char * file_verdict_on PARAMS ((const char *filename));

/*
 * List the style sheets
 */
void list_style_sheets_short PARAMS ((struct a2ps_job * job,
				      FILE * stream));
void list_style_sheets_long PARAMS ((struct a2ps_job * job,
				     FILE * strea));
void list_html_style_sheets PARAMS ((struct a2ps_job * job,
				     FILE * strea));
void list_texinfo_style_sheets PARAMS ((struct a2ps_job * job,
					FILE * strea));

#endif /* not defined _SSHEET_H_ */
